@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005 Universitat d'Alacant / Universidad de Alicante
+ * Copyright (C) 2005--2015 Universitat d'Alacant / Universidad de Alicante
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -27,6 +27,7 @@
 #include <iostream>
 #include <stack>
 #include <apertium/string_utils.h>
+#include "apertium_config.h"
 #include <apertium/unlocked_cstdio.h>
 
 using namespace Apertium;
@@ -61,6 +62,7 @@ Interchunk::Interchunk()
   inword = false;
   null_flush = false;
   internal_null_flush = false;
+  trace = false;
 }
 
 Interchunk::~Interchunk()
@@ -106,10 +108,15 @@ Interchunk::readData(FILE *in)
   me = new MatchExe(t, finals);
  
   // attr_items
+  bool recompile_attrs = Compression::string_read(in) != string(pcre_version());
   for(int i = 0, limit = Compression::multibyte_read(in); i != limit; i++)
   {
     string const cad_k = UtfConverter::toUtf8(Compression::wstring_read(in));
     attr_items[cad_k].read(in);
+    wstring fallback = Compression::wstring_read(in);
+    if(recompile_attrs) {
+      attr_items[cad_k].compile(UtfConverter::toUtf8(fallback));
+    }
   }
 
   // variables
@@ -374,7 +381,7 @@ Interchunk::evalString(xmlNode *element)
   }
   else if(!xmlStrcmp(element->name, (const xmlChar *) "concat"))
   { 
-    string value = "";
+    string value;
     for(xmlNode *i = element->children; i != NULL; i = i->next)
     {
       if(i->type == XML_ELEMENT_NODE)
@@ -384,10 +391,13 @@ Interchunk::evalString(xmlNode *element)
     }
     return value;
   }
-
+  else if(!xmlStrcmp(element->name, (const xmlChar *) "chunk"))
+  {
+    return processChunk(element);
+  }  
   else
   {
-    cerr << "Error: unexpected rvalue expression '" << element->name << endl;
+    cerr << "Error: unexpected rvalue expression '" << element->name << "'" << endl;
     exit(EXIT_FAILURE);
   }
 
@@ -403,7 +413,7 @@ Interchunk::processOut(xmlNode *localroot)
     {
       if(!xmlStrcmp(i->name, (const xmlChar *) "chunk"))
       {
-        processChunk(i);
+        fputws_unlocked(UtfConverter::fromUtf8(processChunk(i)).c_str(), output);
       }
       else // 'b'
       {
@@ -413,41 +423,22 @@ Interchunk::processOut(xmlNode *localroot)
   }
 }
 
-void
+string
 Interchunk::processChunk(xmlNode *localroot)
 {
-  fputwc_unlocked(L'^', output);
+  string result;
+  result.append("^");
   
   for(xmlNode *i = localroot->children; i != NULL; i = i->next)
   {
     if(i->type == XML_ELEMENT_NODE)
     {
-      fputws_unlocked(UtfConverter::fromUtf8(evalString(i)).c_str(), output);
+      result.append(evalString(i));
     }      
   }
   
-  fputwc_unlocked(L'$', output);
-}
-
-void
-Interchunk::processTags(xmlNode *localroot)
-{
-  for(xmlNode *i = localroot->children; i != NULL; i = i->next)
-  {
-    if(i->type == XML_ELEMENT_NODE)
-    {
-      if(!xmlStrcmp(i->name, (xmlChar const *) "tag"))
-      {
-        for(xmlNode *j = i->children; j != NULL; j = j->next)
-        {
-          if(j->type == XML_ELEMENT_NODE)
-          {
-            fputws_unlocked(UtfConverter::fromUtf8(evalString(j)).c_str(), output);
-          }
-        }
-      }
-    }
-  }
+  result.append("$");
+  return result;
 }
 
 void
@@ -461,6 +452,10 @@ Interchunk::processInstruction(xmlNode *localroot)
   {
     processLet(localroot);
   }
+  else if(!xmlStrcmp(localroot->name, (const xmlChar *) "append"))
+  {
+    processAppend(localroot);
+  }
   else if(!xmlStrcmp(localroot->name, (const xmlChar *) "out"))
   {
     processOut(localroot);
@@ -546,6 +541,28 @@ Interchunk::processLet(xmlNode *localroot)
   }
 }
 
+void
+Interchunk::processAppend(xmlNode *localroot)
+{
+  string name;
+  for(xmlAttr *i = localroot->properties; i != NULL; i = i->next)
+  {
+    if(!xmlStrcmp(i->name, (const xmlChar *) "n"))
+    {
+      name = (char *) i->children->content; 
+      break;
+    }
+  }
+
+  for(xmlNode *i = localroot->children; i != NULL; i = i->next)
+  {
+    if(i->type == XML_ELEMENT_NODE)
+    {
+      variables[name].append(evalString(i));
+    }
+  }
+}
+
 void
 Interchunk::processModifyCase(xmlNode *localroot)
 {
@@ -567,7 +584,7 @@ Interchunk::processModifyCase(xmlNode *localroot)
     }
   }
 
-  if(!xmlStrcmp(leftSide->name, (const xmlChar *) "clip"))
+  if(leftSide->name != NULL && !xmlStrcmp(leftSide->name, (const xmlChar *) "clip"))
   {
     int pos = 0;
     xmlChar *part = NULL;
@@ -1168,7 +1185,7 @@ Interchunk::processContainsSubstring(xmlNode *localroot)
 string
 Interchunk::copycase(string const &source_word, string const &target_word)
 {
-  wstring result = L"";
+  wstring result;
   wstring const s_word = UtfConverter::fromUtf8(source_word);
   wstring const t_word = UtfConverter::fromUtf8(target_word);
 
@@ -1279,7 +1296,7 @@ Interchunk::readToken(FILE *in)
     return input_buffer.next();
   }
 
-  wstring content = L"";
+  wstring content;
   while(true)
   {
     int val = fgetwc_unlocked(in);
@@ -1371,6 +1388,12 @@ Interchunk::setNullFlush(bool null_flush)
   this->null_flush = null_flush;
 }
 
+void
+Interchunk::setTrace(bool trace)
+{
+  this->trace = trace;
+}
+
 void
 Interchunk::interchunk_wrapper_null_flush(FILE *in, FILE *out)
 {
@@ -1441,6 +1464,20 @@ Interchunk::interchunk(FILE *in, FILE *out)
     {
       lastrule = rule_map[val-1];      
       last = input_buffer.getPos();
+
+      if(trace)
+      {
+        wcerr << endl << L"apertium-interchunk: Rule " << val << L" ";
+        for (unsigned int ind = 0; ind < tmpword.size(); ind++)
+        {
+          if (ind != 0)
+          {
+            wcerr << L" ";
+          }
+          wcerr << *tmpword[ind];
+        }
+        wcerr << endl;
+      }
     }
 
     TransferToken &current = readToken(in);
@@ -1466,6 +1503,7 @@ Interchunk::interchunk(FILE *in, FILE *out)
 	else
 	{
 	  fputws_unlocked(current.getContent().c_str(), output);
+	  tmpblank.clear();
 	  return;
 	}
 	break;
